import os

from .template import Template
from .errors import TemplateLoadError


# Loads templates from the file system. Assumes files are utf-8 encoded. Compiled templates are
# cached in memory, so they only need to be compiled once. Templates are *not* automatically
# recompiled if the underlying template file changes.
#
# A FileLoader instance should be initialized with a path to a base template directory.
#
#     loader = FileLoader('/path/to/base/directory')
#
# The loader instance can then be called with a filename string. The loader will return the template
# object corresponding the template file or raise a TemplateLoadError if no file can be located.
# Note that the filename string may include subdirectory paths:
#
#     template = loader('foo.txt')
#     template = loader('subdir/foo.txt')
#
class FileLoader:

    def __init__(self, base_dir):
        self.base_dir = base_dir
        self.cache = {}

    def __call__(self, filename):
        if filename in self.cache:
            return self.cache[filename]

        path = os.path.join(self.base_dir, filename)
        if os.path.isfile(path):
            try:
                with open(path, encoding='utf-8') as file:
                    template_string = file.read()
            except OSError as err:
                msg = f"FileLoader cannot load the template file '{filename}'."
                raise TemplateLoadError(msg) from err

            template = Template(template_string, filename)
            self.cache[filename] = template
            return template

        msg = f"FileLoader with base '{self.base_dir}' cannot locate the template file '{filename}'."
        raise TemplateLoadError(msg)


# Loads templates from the file system. Assumes files are utf-8 encoded. Compiled templates are
# cached in memory, so they only need to be compiled once. Templates are automatically recompiled
# if the underlying template file changes.
class FileReloader:

    def __init__(self, base_dir):
        self.base_dir = base_dir
        self.cache = {}

    def __call__(self, filename):
        path = os.path.join(self.base_dir, filename)
        if os.path.isfile(path):
            mtime = os.path.getmtime(path)
            if filename in self.cache:
                if mtime == self.cache[filename][0]:
                    return self.cache[filename][1]

            try:
                with open(path, encoding='utf-8') as file:
                    template_string = file.read()
            except OSError as err:
                msg = f"FileReloader cannot load the template file '{filename}'."
                raise TemplateLoadError(msg) from err

            template = Template(template_string, filename)
            self.cache[filename] = (mtime, template)
            return template

        msg = f"FileReloader with base '{self.base_dir}' cannot locate the template file '{filename}'."
        raise TemplateLoadError(msg)


# Loads templates from a dictionary of template strings. Templates are compiled once and cached for
# future use.
class DictLoader:

    def __init__(self, template_strings):
        self.templates = {}
        self.template_strings = template_strings

    def __call__(self, name):
        if name in self.templates:
            return self.templates[name]
        elif name in self.template_strings:
            template = Template(self.template_strings[name], name)
            self.templates[name] = template
            return template
        msg = f"DictLoader has no entry matching the template name '{name}'."
        raise TemplateLoadError(msg)



import datetime
from . import errors


# User-configurable functions and variables available in all contexts.
builtins = {
    'now': datetime.datetime.now,
    'range': range,
}


# A Context object is a wrapper around the user's input data. Its `.resolve()` method contains
# the lookup-logic for resolving dotted variable names.
class Context:

    def __init__(self, data_dict, template, strict_mode):
        self.stack = []

        # Standard builtins.
        self.stack.append({
            'context': self,
            'is_defined': self.is_defined,
        })

        # User-configurable builtins.
        self.stack.append(builtins)

        # Instance-specific data.
        self.stack.append(data_dict)

        # Nodes can store state information here to avoid threading issues.
        self.stash = {}

        # This reference gives nodes access to their parent template object.
        self.template = template

        # In strict mode undefined variables raise an UndefinedVariable exception.
        self.strict_mode = strict_mode

    def __setitem__(self, key, value):
        self.stack[-1][key] = value

    def __getitem__(self, key):
        for d in reversed(self.stack):
            if key in d:
                return d[key]
        raise KeyError(key)

    def __delitem__(self, key):
        del self.stack[-1][key]

    def __contains__(self, key):
        for d in self.stack:
            if key in d:
                return True
        return False

    def resolve(self, varstring, token):
        # The lists of specific exception types below are important - these are
        # the default access-error types that Python itself throws. We can't use
        # a catch-all except-clause as we need to make sure that custom error types
        # can still bubble up. (A custom error might be thrown by an attempt to access
        # a @property attribute - if we swallow the error here and return Undefined()
        # it can result in a *very* difficult to diagnose bug!)
        words = []
        result = self
        for word in varstring.split('.'):
            words.append(word)
            try:
                result = result[word]
            except (TypeError, AttributeError, KeyError, ValueError):
                try:
                    result = getattr(result, word)
                except (TypeError, AttributeError):
                    if self.strict_mode:
                        msg = f"Cannot resolve the variable '{'.'.join(words)}' "
                        msg += f"in template '{token.template_id}', line {token.line_number}."
                        raise errors.UndefinedVariable(msg, token) from None
                    return Undefined()
        return result

    def is_defined(self, varstring):
        current = self
        for token in varstring.split('.'):
            try:
                current = current[token]
            except:
                try:
                    current = getattr(current, token)
                except:
                    return False
        return True

    def push(self, data=None):
        self.stack.append(data or {})

    def pop(self):
        self.stack.pop()

    def get(self, key, default=None):
        for d in reversed(self.stack):
            if key in d:
                return d[key]
        return default

    def update(self, data):
        self.stack[-1].update(data)


# Null type returned when a context lookup fails.
class Undefined:

    def __str__(self):
        return ''

    def __bool__(self):
        return False

    def __len__(self):
        return 0

    def __contains__(self, key):
        return False

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return True




# Base class for all exception types raised by the template engine.
class TemplateError(Exception):
    pass


# This exception type may be raised while attempting to load a template file.
class TemplateLoadError(TemplateError):
    pass


# This exception type is raised if the lexer cannot tokenize a template string.
class TemplateLexingError(TemplateError):

    def __init__(self, msg, template_id):
        super().__init__(msg)
        self.template_id = template_id


# This exception type may be raised while a template is being compiled.
class TemplateSyntaxError(TemplateError):

    def __init__(self, msg, token):
        super().__init__(msg)
        self.token = token


# This exception type may be raised while a template is being rendered.
class TemplateRenderingError(TemplateError):

    def __init__(self, msg, token):
        super().__init__(msg)
        self.token = token


# This exception type is raised in strict mode if a variable cannot be resolved.
class UndefinedVariable(TemplateError):

    def __init__(self, msg, token):
        super().__init__(msg)
        self.token = token

